###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import IntField, DictField
from onyx.core.depgraph.graph_api import ValueType, CreateInMemory, GetVal
from onyx.core.depgraph.graph_scopes import EvalBlock

from onyx.core.utils.unittest import OnyxTestCase

import unittest

CF_GROWTH = 0.02
G = 0.25   # gearing: D / (D+E)
Kd = 0.06  # cost of debt
Ke = 0.08  # cost of equity
KWDS = {
    "PastCashFlows": {2014: 101.3, 2015: 100.0},
}


###############################################################################
class DCFModel(UfoBase):
    CurrentYear = IntField(default=2016)
    FinalYear = IntField(default=2050)
    PastCashFlows = DictField()
    PastGrowth = DictField()

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def CashFlow(self, graph, year):
        if year < graph(self, "CurrentYear"):
            return graph(self, "PastCashFlows")[year]
        else:
            multiplier = 1.0 + graph(self, "CashFlowGrowth", year)
            return graph(self, "CashFlow", year - 1)*multiplier

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def CashFlowGrowth(self, graph, year):
        if year < graph(self, "CurrentYear"):
            cf = graph(self, "PastCashFlows")
            return cf[year] / cf[year-1] - 1.0
        else:
            return CF_GROWTH

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def WACC(self, graph, year):
        return G*Kd + (1.0 - G)*Ke

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def DiscountFactor(self, graph, year):
        return 1.0 / (1.0 + graph(self, "WACC", year))

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def DCF(self, graph, year):
        if year > graph(self, "FinalYear"):
            return 0.0
        elif year == graph(self, "FinalYear"):
            return graph(self, "CashFlow", year)
        else:
            cf = graph(self, "CashFlow", year)
            df = graph(self, "DiscountFactor", year)
            return cf + df*graph(self, "DCF", year + 1)


###############################################################################
class TestDCFModel(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(DCFModel(Name="model", **KWDS))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_cashflow(self):
        self.assertEqual(GetVal(self.name, "CashFlow", 2016), 102.0)
        self.assertEqual(GetVal(self.name, "CashFlow", 2017), 104.04)

    # -------------------------------------------------------------------------
    def test_cashflow_growth(self):
        cf_2015 = GetVal(self.name, "CashFlowGrowth", 2015)
        self.assertAlmostEqual(cf_2015, -0.0128, 4)
        cf_2016 = GetVal(self.name, "CashFlowGrowth", 2016)
        self.assertEqual(cf_2016, CF_GROWTH)

    # -------------------------------------------------------------------------
    def test_wacc(self):
        self.assertEqual(GetVal(self.name, "WACC", 2017), G*Kd + (1.0 - G)*Ke)

    # -------------------------------------------------------------------------
    def test_discount_factor(self):
        self.assertEqual(GetVal(self.name, "DiscountFactor", 2017),
                         1.0 / (1.0 + G*Kd + (1.0 - G)*Ke))

    # -------------------------------------------------------------------------
    def test_dcf(self):
        def dcf(cf, final_year):
            # --- auxiliary function to calculate result analytically:
            #     dcf = cf * \sum k=0, N (1+g)^k/(1+wacc)^k
            #         = cf * (1 - q^N+1)/(1 - q)
            #     where q = (1+g)/(1+wacc)
            q = (1.0 + CF_GROWTH) / (1.0 + G*Kd + (1.0 - G)*Ke)
            n = final_year - 2016
            return cf*(1.0 - q**(n+1)) / (1.0 - q)

        cf = GetVal(self.name, "CashFlow", 2016)
        fy = GetVal(self.name, "FinalYear")

        self.assertAlmostEqual(GetVal(self.name, "DCF", 2016), dcf(cf, fy), 10)

        # --- test it on shorter cashflows
        with EvalBlock() as eb:
            for fy in (2020, 2030, 2040):
                eb.change_value(self.name, "FinalYear", fy)
                self.assertAlmostEqual(GetVal(self.name,
                                              "DCF", 2016), dcf(cf, fy), 10)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import uuid

__all__ = ["b36encode", "unique_id"]

ALPHABET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'


# -----------------------------------------------------------------------------
def b36encode(hexdigest):
    """
    Description:
        Encode a hexadecimal bynary string in base36. The choice of base 36
        is convenient in that the digits can be represented using the arabic
        numerals 0–9 and the Latin letters A–Z.
        (http://en.wikipedia.org/wiki/Base_36)
    Inputs:
        hexdigest - a hexadecimal bynary string
    Returns:
        A string.
    Example:
        b36encode(hashlib.md5(b"abc-def").hexdigest())
    """
    number = int(hexdigest, 16)

    base36 = ""
    while number:
        number, k = divmod(number, 36)
        base36 = ALPHABET[k] + base36

    return base36 or ALPHABET[0]


# -----------------------------------------------------------------------------
def unique_id(nchars=8):
    """
    Description:
        Return pseudo-unique identifier of nchars alpha-numeric characters
        based on a random UUID.
    Inputs:
        nchars - number of characters in the unique ID (default is 8, which
                 gives 282.e10 distinct possibilities)
    Returns:
        A string.
    """
    if nchars > 25:
        raise ValueError("maximum number of characters is 25")
    buuid = bytes(str(uuid.uuid4()).replace("-", ""), "utf-8")
    return b36encode(buuid)[:nchars]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import StringField

from onyx.core.depgraph.graph_api import ValueType, CreateInMemory
from onyx.core.depgraph.graph_api import GetVal, InvalidateNode, ChildrenSet
from onyx.core.depgraph.graph_scopes import EvalBlock, GraphScope

from onyx.core.utils.unittest import OnyxTestCase

import unittest


###############################################################################
class Diamond(UfoBase):
    A = StringField(default="A")

    # -------------------------------------------------------------------------
    @ValueType()
    def B(self, graph):
        return graph(self, "A") + "B"

    # -------------------------------------------------------------------------
    @ValueType()
    def C(self, graph):
        return graph(self, "A") + "C"

    # -------------------------------------------------------------------------
    @ValueType()
    def D(self, graph):
        return graph(self, "B") + graph(self, "C")


###############################################################################
class TestDiamond(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(Diamond(Name="diamond"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def test_leaves(self):
        self.assertEqual(GetVal(self.name, "A"), "A")

    # -------------------------------------------------------------------------
    def test_errors(self):
        pass
        # self.assertRaises(RuntimeError, GetVal, self.name, "xxx")

    # -------------------------------------------------------------------------
    def test_values(self):
        self.assertEqual(GetVal(self.name, "B"), "AB")
        self.assertEqual(GetVal(self.name, "C"), "AC")
        self.assertEqual(GetVal(self.name, "D"), "ABAC")

    # -------------------------------------------------------------------------
    def test_children(self):
        self.assertEqual(ChildrenSet((self.name, "D"), "A"), {"diamond"})

    # -------------------------------------------------------------------------
    def test_eval_block(self):
        self.test_values()

        with EvalBlock() as eb:
            eb.change_value(self.name, "A", "XXX")
            self.assertEqual(GetVal(self.name, "B"), "XXXB")
            self.assertEqual(GetVal(self.name, "C"), "XXXC")
            self.assertEqual(GetVal(self.name, "D"), "XXXBXXXC")

            InvalidateNode(self.name, "A")

            self.test_values()

            eb.change_value(self.name, "C", "XXX")
            self.assertEqual(GetVal(self.name, "D"), "ABXXX")

        self.test_values()

    # -------------------------------------------------------------------------
    def test_graph_scope(self):
        self.test_values()

        scope = GraphScope()
        scope.change_value(self.name, "A", "XXX")

        self.test_values()

        with scope:
            self.assertEqual(GetVal(self.name, "B"), "XXXB")
            self.assertEqual(GetVal(self.name, "C"), "XXXC")
            self.assertEqual(GetVal(self.name, "D"), "XXXBXXXC")

            InvalidateNode(self.name, "A")

            self.test_values()

            scope.change_value(self.name, "C", "XXX")
            self.assertEqual(GetVal(self.name, "D"), "ABXXX")

        self.test_values()

        with scope:
            self.assertEqual(GetVal(self.name, "D"), "ABXXX")


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import FloatField
from onyx.core.depgraph.graph_api import ValueType
from onyx.core.depgraph.graph_api import GetVal, SetVal, CreateInMemory
from onyx.core.depgraph.graph_api import UseGraph, InvalidateNode
from onyx.core.depgraph.graph_scopes import EvalBlock, GraphScope

from onyx.core.utils.unittest import OnyxTestCase

import unittest


###############################################################################
class test_cls(UfoBase):
    Number = FloatField(default=666.0)

    # -------------------------------------------------------------------------
    @ValueType()
    def Property(self, graph):
        return 1.0 + 2.0*graph(self, "Number")


###############################################################################
class TestGraphScopes(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.instance = CreateInMemory(test_cls(Name="test"))
        self.name = self.instance.Name

    # -------------------------------------------------------------------------
    def tearDown(self):
        super().tearDown()

    # -------------------------------------------------------------------------
    def node_assertEqual(self, obj, vt, value):
        self.assertEqual(GetVal(self.name, vt), value)
        self.assertEqual(GetVal(self.instance, vt), value)

    # -------------------------------------------------------------------------
    def test_evalblock(self):
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- create an eval block
        with EvalBlock() as ev_block:
            self.node_assertEqual(self.name, "Number", 666.0)
            self.node_assertEqual(self.name, "Property", 1333.0)

            # --- now change Number to 0
            ev_block.change_value(self.name, "Number", 0)
            self.node_assertEqual(self.name, "Property", 1.0)

            # --- test multiple changes within nested eval blocks
            with EvalBlock() as new_ev_block:
                new_ev_block.change_value(self.name, "Number", 333)
                self.node_assertEqual(self.name, "Property", 667.0)

                new_ev_block.change_value(self.name, "Number", -333)
                self.node_assertEqual(self.name, "Property", -665.0)

            # --- outside nested eval block
            self.node_assertEqual(self.name, "Property", 1.0)

            # --- now change the calc VT itself
            ev_block.change_value(self.name, "Property", 1976)
            self.node_assertEqual(self.name, "Property", 1976)

            with EvalBlock() as new_ev_block:
                new_ev_block.change_value(self.name, "Property", 333)
                self.node_assertEqual(self.name, "Property", 333)

            # --- outside nested eval block
            self.node_assertEqual(self.name, "Property", 1976)

            # --- invalidating a changed node should take it back to the
            #     original state
            InvalidateNode(self.name, "Property")
            self.node_assertEqual(self.name, "Property", 1.0)

        # --- outside all eval blocks
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

    # -------------------------------------------------------------------------
    def test_graphscope_throwaway(self):
        with GraphScope() as scope:
            self.node_assertEqual(self.name, "Number", 666.0)
            self.node_assertEqual(self.name, "Property", 1333.0)

            scope.change_value(self.name, "Number", 0.0)

            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", 1.0)

            scope.change_value(self.name, "Property", "ABC")

            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", "ABC")

        # --- outside graph scope
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- the graph scope should be usable again
        with scope:
            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", "ABC")

            # --- invalidating a changed node should bring it back to the
            #     original state
            InvalidateNode(self.name, "Property")
            self.node_assertEqual(self.name, "Property", 1.0)

        # --- outside graph scope
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

    # -------------------------------------------------------------------------
    def test_graphscope(self):
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- create a graph scope and set a first change
        scope = GraphScope()
        scope.change_value(self.name, "Number", 0.0)

        # --- the value is changed, but not yet used
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- use the graph scope to enforce existing changes and to add more
        with scope:
            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", 1.0)

            scope.change_value(self.name, "Property", "ABC")

            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", "ABC")

        # --- outside graph scope
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- a graph scope can be reused and all changes are active again
        with scope:
            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", "ABC")

            # --- invalidating a changed node should bring it back to the
            #     original state
            InvalidateNode(self.name, "Property")
            self.node_assertEqual(self.name, "Property", 1.0)

        # --- outside graph scope
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

    # -------------------------------------------------------------------------
    def test_graphscope_switch_graph(self):
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- create a graph scope and set a first change
        scope = GraphScope()
        scope.change_value(self.name, "Number", 0.0)

        self.node_assertEqual(self.name, "Number", 666.0)

        with UseGraph():
            SetVal(self.name, "Number", 5.0)

            self.node_assertEqual(self.name, "Number", 5.0)

            with scope:
                self.node_assertEqual(self.name, "Number", 0.0)
                self.node_assertEqual(self.name, "Property", 1.0)

            self.node_assertEqual(self.name, "Number", 5.0)
            self.node_assertEqual(self.name, "Property", 11.0)

        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

    # -------------------------------------------------------------------------
    def test_nested_graphscopes_1(self):
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- create a graph scope and set a first change
        outer = GraphScope()
        outer.change_value(self.name, "Number", 0.0)

        with outer:
            # --- create a second graph scope within the outer scope so that
            #     inner will fallback on outer and inherit its changes
            inner = GraphScope()

            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", 1.0)

            with inner:
                self.node_assertEqual(self.name, "Number", 0.0)
                self.node_assertEqual(self.name, "Property", 1.0)

                inner.change_value(self.name, "Number", 5)
                self.node_assertEqual(self.name, "Property", 11.0)

                inner.change_value(self.name, "Number", -5)
                self.node_assertEqual(self.name, "Property", -9.0)

            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", 1.0)

        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

    # -------------------------------------------------------------------------
    def test_nested_graphscopes_2(self):
        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)

        # --- create a graph scope and set a first change
        outer = GraphScope()
        outer.change_value(self.name, "Number", 0.0)

        # --- create a second graph scope outside the first scope so that
        #     inner will not fallback on outer and it won't inherit its changes
        inner = GraphScope()

        with outer:
            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", 1.0)

            with inner:
                self.node_assertEqual(self.name, "Number", 666.0)
                self.node_assertEqual(self.name, "Property", 1333.0)

                inner.change_value(self.name, "Number", 5)
                self.node_assertEqual(self.name, "Property", 11.0)

                inner.change_value(self.name, "Number", -5)
                self.node_assertEqual(self.name, "Property", -9.0)

            self.node_assertEqual(self.name, "Number", 0.0)
            self.node_assertEqual(self.name, "Property", 1.0)

        self.node_assertEqual(self.name, "Number", 666.0)
        self.node_assertEqual(self.name, "Property", 1333.0)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
Implements base and meta classes for universal financial objects (UFOs) and all
persistable fields.
"""

from ..datatypes.date import Date

import numpy as np
import getpass
import json
import copy

__all__ = ["FieldError", "UfoError", "UfoBase"]

USER = getpass.getuser()
SPECIAL = ("Name", "ObjType", "Version",
           "ChangedBy", "TimeCreated", "LastUpdated")
SKIP = {"_data", "_json_fields"}.union(SPECIAL)


# -----------------------------------------------------------------------------
def get_base_classes(cls):
    yield cls.__name__
    for base in cls.__bases__:
        yield from get_base_classes(base)


###############################################################################
class custom_encoder(json.JSONEncoder):
    # -------------------------------------------------------------------------
    def default(self, obj):
        if isinstance(obj, Date):
            return {
                "__type__": "Date",
                "year": obj.year,
                "month": obj.month,
                "day": obj.day,
                "hour": obj.hour,
                "minute": obj.minute,
                "second": obj.second,
                "microsecond": obj.microsecond,
            }
        elif isinstance(obj, np.float32):
            return {
                "__type__": "np.float32",
                "value": float(obj),
            }
        elif isinstance(obj, np.float):
            return {
                "__type__": "np.float",
                "value": float(obj),
            }
        elif isinstance(obj, np.int64):
            return {
                "__type__": "np.int64",
                "value": int(obj),
            }
        elif isinstance(obj, np.integer):
            return {
                "__type__": "np.integer",
                "value": int(obj),
            }

        # --- let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


###############################################################################
class custom_decoder(json.JSONDecoder):
    # -------------------------------------------------------------------------
    def __init__(self, *args, **kdws):
        super().__init__(*args, object_hook=self.dict_to_object, **kdws)

    # -------------------------------------------------------------------------
    def dict_to_object(self, obj):
        try:
            typename = obj.pop("__type__")
        except KeyError:
            return obj
        if typename == "Date":
            return Date(**obj)
        elif typename == "np.float":
            return np.float64(obj["value"])
        elif typename == "np.float32":
            return np.float32(obj["value"])
        elif typename == "np.integer":
            return np.int32(obj["value"])
        elif typename == "np.int64":
            return np.int64(obj["value"])
        else:
            # --- put this back together
            obj["__type__"] = typename
            return obj


###############################################################################
class FieldError(Exception):
    pass


###############################################################################
class BaseField(object):
    # -------------------------------------------------------------------------
    def __init__(self, default=None):
        if default is not None:
            self.validate(default)
        self.name = None
        self.default = default

    # -------------------------------------------------------------------------
    def __get__(self, instance, cls):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), "fieldname") works
            return self
        try:
            return instance._data[self.name]
        except KeyError:
            instance._data[self.name] = self.default
            return self.default

    # -------------------------------------------------------------------------
    def __set__(self, instance, val):
        self.validate(val)
        instance._data[self.name] = val

    # -------------------------------------------------------------------------
    def from_json(self, val):
        """
        This method is used to prepare the fild value for json reverse
        serialization (needed for data types that are not naturally supported
        by json format).
        """
        return val

    # -------------------------------------------------------------------------
    def to_json(self, val):
        """
        This method is used to prepare the fild value for json serialization
        (needed for data types that are not naturally supported by json
        format).
        """
        return val

    # -------------------------------------------------------------------------
    def validate(self, val):
        pass

    # -------------------------------------------------------------------------
    def __repr__(self):
        return "{0:s} {1!s}".format(self.__class__.__name__, self.name)


###############################################################################
class MutableField(BaseField):
    pass


###############################################################################
class UfoError(Exception):
    pass


###############################################################################
class UfoMetaClass(type):
    # -------------------------------------------------------------------------
    def __new__(mcl, name, bases, nmspc):
        if "__init__" in nmspc:
            raise UfoError("Classes derived from UfoBase"
                           "cannot implement the __init__ method")

        json_flds = set()

        # --- inherit StoredAttrs of all UFO bases (if any)
        for base in bases:
            if hasattr(base, "_json_fields"):
                json_flds.update(base._json_fields)

        # --- add class specific StoredAttrs and set their name
        for attr_name, attr in nmspc.items():
            if isinstance(attr, BaseField):
                json_flds.add(attr_name)
                attr.name = attr_name

        # --- StoredAttrs is a set with the list of persisted attributes
        nmspc["StoredAttrs"] = json_flds.union(SPECIAL)

        # --- json_flds is a special attribute
        nmspc["_json_fields"] = json_flds

        return super().__new__(mcl, name, bases, nmspc)

    # -------------------------------------------------------------------------
    #  class instantiation
    def __call__(cls, *args, **kwds):
        """
        This method is the one invoked to create a new instance of a class. In
        turn, it invokes the __new__ method of the associated class.
        NB: this method will not be invoked when an object is de-serialized.
        """
        if len(args):
            raise UfoError("Classes derived from UfoBase cannot "
                           "be instantiated with positional arguments")

        # --- create an instance of the class
        instance = cls.__new__(cls)

        # --- set default values for the special attributes
        now = Date.now()

        setattr(instance, "Name", kwds.pop("Name", None))
        setattr(instance, "ObjType", kwds.pop("ObjType", cls.__name__))
        setattr(instance, "Version", kwds.pop("Version", 0))
        setattr(instance, "ChangedBy", kwds.pop("ChangedBy", USER))
        setattr(instance, "TimeCreated", kwds.pop("TimeCreated", now))
        setattr(instance, "LastUpdated", kwds.pop("LastUpdated", now))

        # --- set stored attributes based on input values
        for attr, val in kwds.items():
            if attr in instance._json_fields:
                setattr(instance, attr, val)
            else:
                raise UfoError("Unrecognized stored "
                               "attribute for {0!s}: {1:s}".format(cls, attr))

        # --- call post-initialization method if available
        if hasattr(cls, "__post_init__"):
            instance.__post_init__()

        return instance


###############################################################################
class UfoBase(metaclass=UfoMetaClass):
    # -------------------------------------------------------------------------
    def __new__(cls, *args, **kwds):
        # --- the instance is created using the super-class __new__ method
        instance = super().__new__(cls, *args, **kwds)

        # --- the _data attribute, which will store the values of all the
        #     stored attributes, is initialized here to an empty dictionary.
        instance._data = {}

        # --- for each mutable field with non-trivial default value, make a
        #     deepcopy of such default value to avoid it being shared across
        #     multiple instances.
        for field_name in instance._json_fields:
            field = getattr(cls, field_name)
            if isinstance(field, MutableField) and field.default is not None:
                setattr(instance, field_name, copy.deepcopy(field.default))

        return instance

    # -------------------------------------------------------------------------
    def to_json(self):
        if self.Name == "":
            raise UfoError("Name attribute cannot be an empty string")

        cls = self.__class__

        # --- create json object
        json_data = {name: getattr(cls, name).to_json(getattr(self, name))
                     for name in self._json_fields}

        return json.dumps(json_data, cls=custom_encoder)

    # -------------------------------------------------------------------------
    def from_json(self, values):
        cls = self.__class__
        for name, value in json.loads(values, cls=custom_decoder).items():
            self._data[name] = getattr(cls, name).from_json(value)

    # -------------------------------------------------------------------------
    def __eq__(self, other):
        self_attrs = self.StoredAttrs.difference(SKIP)
        other_attrs = other.StoredAttrs.difference(SKIP)

        if self_attrs != other_attrs:
            return False

        for attr in self_attrs:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    # -------------------------------------------------------------------------
    def __repr__(self):
        return "<{0:s} {1!s}>".format(self.ObjType, self.Name)

    # -------------------------------------------------------------------------
    def clone(self, **overrides):
        """
        Create a copy of current object and apply optional overrides.
        """
        # --- create timestamp
        now = Date.now()

        # --- create a new instance from current stored attributes and
        #     overwrite the special attributes. Finally, overrides are applied.
        #     NB: attributes with None as value are excluded as it is assumed
        #         they weren't set explicitly.
        stored = {attr: copy.deepcopy(getattr(self, attr))
                  for attr in self.StoredAttrs}
        stored = {attr: val for attr, val in stored.items() if val is not None}
        stored.update({
            "Name": None,
            "Version": 0,
            "TimeCreated": now,
            "LastUpdated": now,
            "ChangedBy": USER.upper(),
        })
        stored.update(overrides)

        return self.__class__(**stored)

    # -------------------------------------------------------------------------
    def copy_from(self, other):
        """
        Overwrite all attributes with those of another instance.
        """
        if self.ObjType != other.ObjType:
            raise TypeError("Can only copy from an object of the same type")
        for attr in other.StoredAttrs:
            if attr in SKIP:
                continue
            setattr(self, attr, getattr(other, attr))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
This module implements a Dependency Graph that works on top of the ObjDb
object database
"""

from .. import database as onyx_db
from .. import depgraph as onyx_dg

import functools
import collections
import weakref
import copy

__all__ = ["GraphError"]


###############################################################################
class GraphError(Exception):
    pass


###############################################################################
class DependencyGraph(collections.UserDict):
    """
    Class representing the onyx Dependency Graph (a Directed Acyclic Graph).
    Current implementation uses a dictionary that maps (ObjectName, VT, args)
    tuples to graph nodes (instances of GraphNode class).
    """
    # -------------------------------------------------------------------------
    def __init__(self):
        super().__init__()
        # --- for each object, we keep a mapping of VTs by instance for the VTs
        #     that have been visited while building the graph
        self.vts_by_obj = dict()

    # -------------------------------------------------------------------------
    def clear(self):
        super().clear()
        self.vts_by_obj.clear()

    # -------------------------------------------------------------------------
    def __deepcopy__(self, memo):
        clone = self.__new__(self.__class__)
        clone.data = self.data.copy()
        clone.vts_by_obj = self.vts_by_obj.copy()
        memo[id(self)] = clone
        return clone


###############################################################################
class GraphNode(object):
    """
    Base class representing a node in the Dependency Graph. Implements methods
    to get the node value (using memoization), to invalidate it (when
    recalculation is needed), and to clone the node (for changing its value).
    """
    __slots__ = ("VT", "obj_ref", "parents", "children", "valid", "value")

    # -------------------------------------------------------------------------
    def __init__(self, VT, obj_ref):
        self.VT = VT
        self.obj_ref = obj_ref
        self.parents = set()
        self.children = set()
        self.valid = False
        self.value = None

    # -------------------------------------------------------------------------
    def get_id(self):
        return self.obj_ref().Name, self.VT, ()

    # -------------------------------------------------------------------------
    def get_value(self):
        if self.valid:
            return self.value
        else:
            # --- before calculating the current value of a node we clean-up
            #     the children set: this is needed to make sure that, once the
            #     node is invalidated, we don't keep references to spurious
            #     nodes.
            #     This is achieved by discarding this node from the set of
            #     parents of all its children
            node_id = self.get_id()
            for child in self.children:
                onyx_dg.graph[child].parents.discard(node_id)
            self.children.clear()
            # --- get the ValueType's value from the instance itself and return
            self.value = getattr(self.obj_ref(), self.VT)
            self.valid = True
            return self.value

    # -------------------------------------------------------------------------
    def clone(self):
        cls = self.__class__
        new_node = cls.__new__(cls)
        new_node.VT = self.VT
        new_node.obj_ref = self.obj_ref
        new_node.parents = self.parents.copy()
        new_node.children = self.children.copy()
        new_node.valid = self.valid
        # --- value can be an instance of a mutable class, make a proper copy.
        new_node.value = copy.deepcopy(self.value)
        return new_node

    # -------------------------------------------------------------------------
    def invalidate(self):
        # --- recursively invalidate all parents
        for parent in self.parents:
            onyx_dg.graph[parent].invalidate()

        # --- invalidate this node
        self.valid = False


###############################################################################
class PropertyNode(GraphNode):
    """
    Derived class used Property ValueTypes.
    """
    __slots__ = ()


###############################################################################
class SettableNode(GraphNode):
    """
    Derived class used for nodes whose value can be explicitly set (such as
    StoredAttrs and descriptors that implement a setter).
    """
    __slots__ = ()


###############################################################################
class CallableNode(GraphNode):
    __slots__ = ("args",)

    # -------------------------------------------------------------------------
    def __init__(self, VT, obj_ref, args):
        super().__init__(VT, obj_ref)
        self.args = args

    # -------------------------------------------------------------------------
    def get_id(self):
        return self.obj_ref().Name, self.VT, self.args

    # -------------------------------------------------------------------------
    def get_value(self, *args):
        assert args == self.args
        if self.valid:
            return self.value
        else:
            # --- before calculating the current value of a node we clean-up
            #     the children set: this is needed to make sure that, once the
            #     node is invalidated, we don't keep references to spurious
            #     nodes.
            #     This is achieved by discarding this node from the set of
            #     parents of all its children
            node_id = self.get_id()
            for child in self.children:
                onyx_dg.graph[child].parents.discard(node_id)
            self.children.clear()
            # --- call the object's method with the arguments for this node
            self.value = getattr(self.obj_ref(), self.VT)(*args)
            self.valid = True
            return self.value

    # -------------------------------------------------------------------------
    def clone(self):
        new_node = super().clone()
        # --- shallow copies here should be fine
        new_node.args = self.args[:]
        return new_node


###############################################################################
class PropSubGraphNode(GraphNode):
    __slots__ = ("kwds",)

    # -------------------------------------------------------------------------
    def __init__(self, VT, obj_ref):
        super().__init__(VT, obj_ref)
        self.kwds = {}

    # -------------------------------------------------------------------------
    def get_value(self, **kwds):
        # --- if args and kwds don't match stored values, the graph node is
        #     assumed no longer valid
        if kwds != self.kwds:
            self.valid = False
            self.kwds = kwds

        if self.valid:
            return self.value
        else:
            # --- before calculating the current value of a node we clean-up
            #     the children set: this is needed to make sure that, once the
            #     node is invalidated, we don't keep references to spurious
            #     nodes.
            #     This is achieved by discarding this node from the set of
            #     parents of all its children
            node_id = self.get_id()
            for child in self.children:
                onyx_dg.graph[child].parents.discard(node_id)
            self.children.clear()
            # --- call the object's method with provided arguments
            self.value = getattr(self.obj_ref(), self.VT)(**kwds)
            self.valid = True
            return self.value

    # -------------------------------------------------------------------------
    def clone(self):
        new_node = super().clone()
        # --- shallow copies here should be fine
        new_node.kwds = self.kwds.copy()
        return new_node


###############################################################################
class BaseVT(object):
    pass


###############################################################################
class PropertyVT(BaseVT):
    def __init__(self, fget):
        # --- ensure that the decorated method has the following signature:
        #         f(self, graph)
        #
        if fget.__code__.co_argcount == 2:
            self.fget = fget
        elif fget.__code__.co_argcount < 2:
            raise GraphError("Missing self and/or "
                             "graph in the method definition")
        elif fget.__code__.co_argcount > 2:
            raise GraphError("A Property VT cannot accept "
                             "extra arguments besides self and graph")

    def __get__(self, instance, cls=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), "vt") works fine
            return self

        caller = (instance.Name, self.fget.__name__, ())
        return self.fget(instance, functools.partial(get_node_val, caller))

    def __set__(self, instance, value):
        raise AttributeError("Cannot set a Property VT")

    def node(self, VT, ref, args):
        return PropertyNode(VT, ref)


###############################################################################
class SettableVT(BaseVT):
    def __init__(self, fget, fset, name):
        self.__name__ = name
        # ---ensure that getter and setter functions have the following
        #     signature:
        #         fget(self, graph), fset(self, graph, value)
        #
        if fget.__code__.co_argcount == 2:
            self.fget = fget
        elif fget.__code__.co_argcount < 2:
            raise GraphError("Missing self and/or "
                             "graph in the getter definition")
        elif fget.__code__.co_argcount > 2:
            raise GraphError("A Settable VT getter cannot accept "
                             "extra arguments besides self and graph")
        if fset.__code__.co_argcount == 3:
            self.fset = fset
        elif fset.__code__.co_argcount < 3:
            raise GraphError("Missing self and/or graph "
                             "and/or value in the setter definition")
        elif fset.__code__.co_argcount > 3:
            raise GraphError("A Settable VT setter cannot accept "
                             "extra arguments besides self, graph, and value")

    def __get__(self, instance, cls=None):
        if instance is None:
            # --- it's convenient to return the descriptor itself when accessed
            #     on the class so that getattr(type(obj)), "vt") works fine
            return self
        caller = (instance.Name, self.__name__, ())
        return self.fget(instance, functools.partial(get_node_val, caller))

    def __set__(self, instance, value):
        caller = (instance.Name, self.__name__, ())
        self.fset(instance, functools.partial(get_node_val, caller), value)

    def node(self, VT, ref, args):
        return SettableNode(VT, ref)


###############################################################################
class CallableVT(BaseVT):
    """
    Description:
        Descriptor used for callable ValueTypes (methods with input arguments).
        Notice that for a callable ValueType the input arguments are part of
        the node_id so that the same ValueType called with different arguments
        will produce different nodes in the graph.
        The limitation is that only positional arguments are supported.
    Signature, example:
        func(self, graph, x, y, z=1, ...)
    """
    def __init__(self, func):
        if func.__code__.co_argcount < 3:
            raise GraphError("CallableVT signature requires a "
                             "minimum of one argument besides self and graph")
        self.func = func

    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    def __set__(self, instance, value):
        raise AttributeError("Cannot set a CalcPositional VT")

    def __call__(self, *args):
        caller = (self.instance.Name, self.func.__name__, args)
        return self.func(self.instance,
                         functools.partial(get_node_val, caller), *args)

    def node(self, VT, ref, args):
        return CallableNode(VT, ref, args)


###############################################################################
class PropSubGraphVT(BaseVT):
    """
    Description:
        Descriptor used to represent a subgraph with property-like nodes.
        The input arguments are treated as children of the node so that when
        called with different arguments the node in the graph will always be
        invalidated.
        The limitation is that only keyword arguments are supported.
    Signature, example:
        func(self, graph, x=1, y=None, ...)
    """
    def __init__(self, func):
        argcount = func.__code__.co_argcount
        if argcount < 3:
            raise GraphError("PropSubGraphVT signature requires a "
                             "minimum of one argument besides self and graph")
        if not func.__defaults__ or  (argcount - 2 != len(func.__defaults__)):
            raise GraphError("PropSubGraphVT signature "
                             "requires all arguments besides self "
                             "and graph to provide a default value")
        self.func = func

    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    def __set__(self, instance, value):
        raise AttributeError("Cannot set a CalcKeywords VT")

    def __call__(self, **kwds):
        caller = (self.instance.Name, self.func.__name__, ())
        graph = functools.partial(get_node_val, caller)
        return self.func(self.instance, graph, **kwds)

    def node(self, VT, ref, args):
        return PropSubGraphNode(VT, ref)


# -----------------------------------------------------------------------------
def create_node(node_id):
    """
    Description:
        Helper function used to create a new node.
    Inputs:
        node_id - the (Name, VT, args) tuple
    Returns:
        A subclass of GraphNode.
    """
    if len(node_id) == 2:
        obj_name, VT = node_id
        args = ()
    else:
        obj_name, VT, args = node_id

    # --- the instance of the graph keeps a map of VTs by object
    try:
        onyx_dg.graph.vts_by_obj[obj_name].add((VT, args))
    except KeyError:
        onyx_dg.graph.vts_by_obj[obj_name] = {(VT, args)}

    obj = onyx_db.obj_clt.get(obj_name)

    if VT == "StoredAttrs":
        # --- StoredAttrs is special insofar that we don't want it to be
        #     settable/changeable
        return GraphNode(VT, weakref.ref(obj))
    elif VT in obj.StoredAttrs:
        # --- all stored attributes as well as StoredAttrs itself are settable
        #     by definition
        return SettableNode(VT, weakref.ref(obj))
    else:
        # --- use the node constructor of the ValueType descriptor itself
        return getattr(obj.__class__, VT).node(VT, weakref.ref(obj), args)


# -----------------------------------------------------------------------------
def get_node_val(caller, obj, VT, *args, **kwds):
    """
    Description:
        This is the proper way of calling a VT on-graph. Use this function to
        get a value type from within a method that has been decorated by
        @ValueType().
    Inputs:
        caller - caller node identity
        obj    - target object as defined in database (or memory)
        VT     - target value type (a stored attribute or a method of an ufo
                 class decorated by @ValueType())
        *args  - positional arguments used to call the target method
        **kwds -      named arguments used to call the target method
    Returns:
        The VT value.
    """
    # --- get the calling node from the graph (the node is created if it
    #     doesn't exsist)
    try:
        caller_node = onyx_dg.graph[caller]
    except KeyError:
        onyx_dg.graph[caller] = caller_node = create_node(caller)

    # --- define target id tuple
    if isinstance(obj, str):
        target = (obj, VT, args)
    else:
        target = (obj.Name, VT, args)

    # --- get the target node from the graph (the node is created if it
    #     doesn't exsist)
    #     NB: this raises the following exceptions BEFORE changing the state
    #         of the graph:
    #         - ObjNotFound    if obj is not found in database or memory
    #         - AttributeError if obj doesn't have a VT attribute
    try:
        target_node = onyx_dg.graph[target]
    except KeyError:
        onyx_dg.graph[target] = target_node = create_node(target)

    caller_node.children.add(target)
    target_node.parents.add(caller)

    # --- now that the graph is set up, return target node value. we make a
    #     deepcopy to make sure we never return a reference to a mutable value
    return copy.deepcopy(target_node.get_value(*args, **kwds))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..table import Table

import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- perform set-up actions, if any
        self.list = [
            ("foo", "boo", "goo"),
            (3.0, "a", 5),
            (1.0, "b", 6),
            (2.0, "c", 7),
            (2.0, "d", 9),
        ]
        self.table = Table(self.list)

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_to_list(self):
        # --- NB: for proper comparison, each row of the table array must
        #         be converted to a tuple
        self.assertEqual([tuple(r) for r in self.table.to_list()], self.list)

    def test_len(self):
        self.assertEqual(len(self.table), len(self.list)-1)

    def test_getitem(self):
        self.assertEqual(self.table[1].goo, 6)
        self.assertEqual(self.table[1]["goo"], 6)

    def test_append(self):
        newrow = (4.0, "d", 8)
        self.list.append(newrow)
        self.table.append(newrow)
        self.assertEqual([tuple(r) for r in self.table.to_list()], self.list)

    def test_column(self):
        self.assertEqual(self.table.column("boo"), ["a", "b", "c", "d"])

    def test_find(self):
        self.assertEqual([row.values for row in self.table.find("foo", 2.0)],
                         [[2.0, "c", 7], [2.0, "d", 9]])

    def test_pickling(self):
        new_table = pickle.loads(pickle.dumps(self.table, 2))
        # --- check that the two tables are identical
        self.assertEqual(new_table, self.table)
        # --- check that data representations are identical
        self.assertEqual([tuple(r) for r in new_table.to_list()], self.list)
        # --- check that string representations are identical
        self.assertEqual(new_table.__str__(), self.table.__str__())

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ...datatypes.date import Date, DateError
from ..date_fns import LYY2Date, Date2LYY, Date2QQYY
from ..date_fns import CalcTerm, DateRange, CountBizDays

import unittest


# --- unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- perform set-up actions, if any
        self.lyy_codes = ["F08", "G08", "H08", "J08", "K08", "M08",
                          "N08", "Q08", "U08", "V08", "X08", "Z08"]
        self.qqyy_codes = ["Q108", "Q108", "Q108", "Q208", "Q208", "Q208",
                           "Q308", "Q308", "Q308", "Q408", "Q408", "Q408"]
        self.dates = [Date.parse(lyy) for lyy in self.lyy_codes]

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_LYY2Date(self):
        self.assertRaises(DateError, LYY2Date, "W99")
        for lyy, d in zip(self.lyy_codes, self.dates):
            self.assertEqual(LYY2Date(lyy), d)

    def test_Date2LYY(self):
        self.assertRaises(DateError, Date2LYY, Date(1900, 1, 1))
        self.assertRaises(DateError, Date2LYY, Date(2100, 1, 1))
        for lyy, d in zip(self.lyy_codes, self.dates):
            self.assertEqual(Date2LYY(d), lyy)

    def test_Date2QQYY(self):
        self.assertRaises(DateError, Date2QQYY, Date(1900, 1, 1))
        self.assertRaises(DateError, Date2QQYY, Date(2100, 1, 1))
        for qqyy, d in zip(self.qqyy_codes, self.dates):
            self.assertEqual(Date2QQYY(d), qqyy)

    def test_CalcTerm(self):
        self.assertEqual(CalcTerm(Date(2006, 12, 31), Date(2007, 12, 31)), 1.0)
        # --- 2008 is a leap year
        self.assertEqual(CalcTerm(Date(2007, 12, 31),
                                  Date(2008, 12, 31), 366), 1.0)

    def test_DateRange(self):
        self.assertEqual(self.dates,
                         [d for d in DateRange(Date(2008, 1, 1),
                                               Date(2008, 12, 31), "+1m")])

    def test_CountBizDays(self):
        self.assertEqual(261, CountBizDays(Date(2007, 1, 1),
                                           Date(2007, 12, 31)))
        self.assertEqual(262, CountBizDays(Date(2008, 1, 1),
                                           Date(2008, 12, 31)))

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..hlocv import HlocvCurve
from ..date import Date

import numpy as np
import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- set the seed to get predictable values
        np.random.seed(666)

        # --- generate random values
        x = np.cumsum(np.random.standard_normal(12+1))

        # --- create a test curve
        dates = [Date.parse(d)
                 for d in ["F08", "G08", "H08", "J08", "K08", "M08",
                           "N08", "Q08", "U08", "V08", "X08", "Z08"]]

        xh = lambda k: max(x[k], x[k+1])  # analysis:ignore
        xl = lambda k: min(x[k], x[k+1])  # analysis:ignore

        values = [(xh(k), xl(k), x[k], x[k+1], 100) for k in range(len(dates))]
        self.refcrv = HlocvCurve(dates, values)

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_set(self):
        value = [1.0, 2.0, 3.0, 4.0, 5.0]
        # --- set at the front of the curve
        date = Date(2007, 1, 1)
        self.refcrv[date] = value
        self.assertTrue(np.allclose(self.refcrv[date], value))
        knot = self.refcrv.front
        self.assertEqual(knot.date, date)
        self.assertTrue(np.allclose(knot.value, value))

        # --- set an existing knot
        date = Date(2008, 6, 1)
        self.assertTrue(np.allclose(self.refcrv[date],
                                    np.array([2.72447971, 2.70545145,
                                              2.70545145, 2.72447971, 100.0])))
        self.refcrv[date] = value
        self.assertTrue(np.allclose(self.refcrv[date], value))

        # --- set missing knot in the middle
        date = Date(2008, 6, 15)
        self.assertRaises(IndexError, self.refcrv.__getitem__, date)
        self.refcrv[date] = value
        self.assertTrue(np.allclose(self.refcrv[date], value))

        # --- set at the back of the curve
        date = Date.today()
        self.refcrv[date] = value
        self.assertTrue(np.allclose(self.refcrv[date], value))
        knot = self.refcrv.back
        self.assertEqual(knot.date, date)
        self.assertTrue(np.allclose(knot.value, value))

    def test_del(self):
        # --- delete an existing knot
        del self.refcrv[Date(2008, 6, 1)]
        self.assertFalse(Date(2008, 6, 1) in self.refcrv)
        # --- delete a missing knot
        self.assertRaises(IndexError,
                          self.refcrv.__delitem__, Date(2008, 6, 1))
        self.assertRaises(IndexError,
                          self.refcrv.__delitem__, Date(2009, 6, 1))

    def test_has(self):
        self.assertFalse(Date(2007, 6, 1) in self.refcrv)
        self.assertTrue(Date(2008, 6, 1) in self.refcrv)
        self.assertFalse(Date(2008, 6, 15) in self.refcrv)
        self.assertFalse(Date(2009, 6, 1) in self.refcrv)

    def test_back_and_front(self):
        self.assertEqual(self.refcrv.front.date, Date(2008, 1, 1))
        self.assertAlmostEqual(self.refcrv.front.value[0], 1.30415409, 8)
        self.assertEqual(self.refcrv.back.date, Date(2008, 12, 1))
        self.assertAlmostEqual(self.refcrv.back.value[0], 2.29092902, 8)

    def test_pickling(self):
        new_crv = pickle.loads(pickle.dumps(self.refcrv, 2))
        self.assertEqual(new_crv, self.refcrv)
        self.assertEqual(new_crv.__str__(), self.refcrv.__str__())

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..database.objdb import ObjNotFound
from ..database.objdb_api import GetObj, AddObj, UseDatabase
from ..database.tsdb_api import TsDbUseDatabase
from ..depgraph.graph_api import CreateInMemory
from .startup import OnyxInit, load_system_configuration
from .base36 import unique_id

from ..database.objdb import ObjDbClient
from ..database.tsdb import TsDbClient

import psycopg2
import psycopg2.extensions as psycopg2_ext
from psycopg2 import sql

import unittest

__all__ = [
    "AddIfMissing",
    "UseEphemeralDbs",
    "ObjDbTestCase",
    "TsDbTestCase",
    "OnyxTestCase",
]

CREATE_CMD = "CREATE DATABASE {};"
DROP_CMD = "DROP DATABASE {};"
TERMINATE_OTHER_CONNECTIONS = """
SELECT pg_terminate_backend(pg_stat_activity.pid)
FROM pg_stat_activity
WHERE pg_stat_activity.datname = %s
  AND pid <> pg_backend_pid();
"""

# -----------------------------------------------------------------------------
def AddIfMissing(obj, in_memory=False):
    try:
        return GetObj(obj.Name)
    except ObjNotFound:
        if in_memory:
            return CreateInMemory(obj)
        else:
            return AddObj(obj)


###############################################################################
class UseEphemeralDbs():
    # -------------------------------------------------------------------------
    def __init__(self):
        config = load_system_configuration()
        uuid = unique_id()
        self.user = config.get("test", "user")
        self.objdb = "{0:s}_objdb_{1:s}".format(self.user, uuid).lower()
        self.tsdb = "{0:s}_tsdb_{1:s}".format(self.user, uuid).lower()
        config.set("test", "objdb", self.objdb)
        config.set("test", "tsdb", self.tsdb)
        self.conn = psycopg2.connect(dbname="postgres", user=self.user)
        self.conn.set_isolation_level(psycopg2_ext.ISOLATION_LEVEL_AUTOCOMMIT)

    # -------------------------------------------------------------------------
    def __enter__(self):
        curs = self.conn.cursor()
        curs.execute(sql.SQL(CREATE_CMD).format(sql.Identifier(self.objdb)))
        curs.execute(sql.SQL(CREATE_CMD).format(sql.Identifier(self.tsdb)))
        return self

    # -------------------------------------------------------------------------
    def __exit__(self, *args, **kwds):
        curs = self.conn.cursor()
        curs.execute(TERMINATE_OTHER_CONNECTIONS, (self.objdb,))
        curs.execute(sql.SQL(DROP_CMD).format(sql.Identifier(self.objdb)))
        curs.execute(TERMINATE_OTHER_CONNECTIONS, (self.tsdb,))
        curs.execute(sql.SQL(DROP_CMD).format(sql.Identifier(self.tsdb)))


###############################################################################
class ObjDbTestCase(unittest.TestCase):
    # -------------------------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        config = load_system_configuration()
        objdb = config.get("test", "objdb")
        user = config.get("test", "user")

        if objdb == config.get("database", "objdb"):
            raise RuntimeError("Trying to run unittests on "
                               "production database ObjDb={0:s}".format(objdb))

        cls.objdb = objdb
        cls.user = user
        cls.clt = ObjDbClient(objdb, user, check=False)

        # --- create tables and functions
        cls.clt.initialize()

    # -------------------------------------------------------------------------
    @classmethod
    def tearDownClass(cls):
        # --- cleanup tables and functions
        cls.clt.cleanup()
        cls.clt.close()

    # -------------------------------------------------------------------------
    def setUp(self):
        self.context = UseDatabase(database=self.objdb, user=self.user)
        self.context.__enter__()

    # -------------------------------------------------------------------------
    def tearDown(self):
        if hasattr(self, "context"):
            self.context.__exit__(None, None, None)


###############################################################################
class TsDbTestCase(unittest.TestCase):
    # -------------------------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        config = load_system_configuration()
        tsdb = config.get("test", "tsdb")
        user = config.get("test", "user")

        if tsdb == config.get("database", "tsdb"):
            raise RuntimeError("Trying to run unittests on "
                               "production database TsDb={0:s}".format(tsdb))

        cls.tsdb = tsdb
        cls.user = user
        cls.clt = TsDbClient(tsdb, user, check=False)

        # --- create tables and functions
        cls.clt.initialize()

    # -------------------------------------------------------------------------
    @classmethod
    def tearDownClass(cls):
        # --- cleanup tables and functions
        cls.clt.cleanup()
        cls.clt.close()

    # -------------------------------------------------------------------------
    def setUp(self):
        self.context = TsDbUseDatabase(database=self.tsdb, user=self.user)
        self.context.__enter__()

    # -------------------------------------------------------------------------
    def tearDown(self):
        if hasattr(self, "context"):
            self.context.__exit__(None, None, None)


###############################################################################
class OnyxTestCase(unittest.TestCase):
    # -------------------------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        config = load_system_configuration()
        objdb = config.get("test", "objdb")
        tsdb = config.get("test", "tsdb")
        user = config.get("test", "user")
        host = config.get("test", "host", fallback=None)

        if objdb == config.get("database", "objdb"):
            raise RuntimeError("Trying to run unittests on "
                               "production database ObjDb={0:s}".format(objdb))
        if tsdb == config.get("database", "tsdb"):
            raise RuntimeError("Trying to run unittests on "
                               "production database TsDb={0:s}".format(tsdb))

        cls.obj_clt = ObjDbClient(objdb, user, host, check=False)
        cls.ts_clt = TsDbClient(tsdb, user, host, check=False)

        cls.objdb = objdb
        cls.tsdb = tsdb
        cls.user = user
        cls.host = host

        # --- create tables and functions
        cls.obj_clt.initialize()
        cls.ts_clt.initialize()

    # -------------------------------------------------------------------------
    @classmethod
    def tearDownClass(cls):
        # --- cleanup tables and functions
        cls.obj_clt.cleanup()
        cls.obj_clt.close()
        cls.ts_clt.cleanup()
        cls.ts_clt.close()

    # -------------------------------------------------------------------------
    def setUp(self):
        self.context = OnyxInit(objdb=self.objdb,
                                tsdb=self.tsdb, user=self.user, host=self.host)
        self.context.__enter__()

    # -------------------------------------------------------------------------
    def tearDown(self):
        if hasattr(self, "context"):
            self.context.__exit__(None, None, None)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .gcurve import GCurve

import numpy as np

__all__ = ["Curve"]


# -----------------------------------------------------------------------------
def is_numlike(obj):
    """
    Helper function used to determine if an object behaves like a number.
    """
    try:
        obj + 1.0
    except TypeError:
        return False
    else:
        return True


###############################################################################
class Curve(GCurve):
    """
    Curve class based on GCurve and supporting numerical values only.
    """
    # --- this is the default datatype for a Curve
    dtype = np.float64

    # -------------------------------------------------------------------------
    #  A few methods for simple curve algebra. It's always assumed that the
    #  two curves are already aligned
    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self.create_raw(self.dates, self.values + other.values)
        else:
            return self.create_raw(self.dates, self.values + other)

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            return self.create_raw(self.dates, self.values - other.values)
        else:
            return self.create_raw(self.dates, self.values - other)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            return self.create_raw(self.dates, self.values * other.values)
        elif is_numlike(other):
            return self.create_raw(self.dates, self.values * other)
        else:
            TypeError("cannot multiply a {0:s} "
                      "by a {1:s}".format(self.__class__, other.__class__))

    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            return self.create_raw(self.dates, self.values / other.values)
        elif is_numlike(other):
            return self.create_raw(self.dates, self.values / other)
        else:
            TypeError("cannot multiply a {0:s} "
                      "by a {1:s}".format(self.__class__, other.__class__))

    def __radd__(self, scalar):
        return self.create_raw(self.dates, scalar + self.values)

    def __rsub__(self, scalar):
        return self.create_raw(self.dates, scalar - self.values)

    def __rmul__(self, scalar):
        return self.create_raw(self.dates, scalar * self.values)

    def __rtruediv__(self, scalar):
        return self.create_raw(self.dates, scalar / self.values)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ...datatypes.date import Date
from ...datatypes.curve import Curve
from ..ufo_base import UfoBase
from ..ufo_fields import DateField, SetField, CurveField

__all__ = []

DATES = [Date(2012, 8, 6), Date(2014, 11, 28)]


class ufocls(UfoBase):
    Birthday = DateField(default=Date(1977, 6, 8))
    OtherDates = SetField(default=set(DATES))
    SimpleCurve = CurveField(default=Curve(dates=DATES, values=[1, 2]))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .gcurve import GCurve, CurveError

import numpy as np
import enum

__all__ = ["HlocvCurve", "HlocvCurveError"]


###############################################################################
class Fields(enum.IntEnum):
    High = 0
    Low = 1
    Open = 2
    Close = 3
    Volume = 4


# -----------------------------------------------------------------------------
class HlocvCurveError(CurveError):
    pass


###############################################################################
class HlocvCurve(GCurve):
    """
    Class representing a HLOCV curve, generally used for daily stock prices. It
    inherits from the generic GCurve.
    """
    # --- this is the default datatype for a HlocvCurve
    dtype = np.float64

    # -------------------------------------------------------------------------
    def __init__(self, dates=None, values=None):
        if values is not None and len(values):
            _, nfields = np.array(values).shape
            if nfields != len(Fields):
                raise HlocvCurveError("values should be an "
                                      "array/list of n rows "
                                      "and 5 columns (one for each field)")

        super().__init__(dates, values)

    def __setitem__(self, date, val):
        val = np.array(val, dtype=self.dtype)
        idx = self.dates.searchsorted(date)
        if idx == len(self.dates):
            self.dates = np.append(self.dates, date)
            self.values = np.vstack([self.values, val])
        elif self.dates[idx] == date:
            self.values[idx,:] = val  # analysis:ignore
        else:
            self.dates = np.insert(self.dates, idx, date)
            self.values = np.insert(self.values, idx, val, axis=0)

    def __delitem__(self, date):
        idx = self.dates.searchsorted(date)
        if self.dates[idx] == date:
            self.dates = np.delete(self.dates, idx)
            self.values = np.delete(self.values, idx, axis=0)
        else:
            raise IndexError

    # -------------------------------------------------------------------------
    #  methods to return arrays for a specific HLOCV field:
    #    - highs
    #    - lows
    #    - opens
    #    - closes
    #    - volumes
    @property
    def highs(self):
        return self.values[:,Fields.High]  # analysis:ignore

    @property
    def lows(self):
        return self.values[:,Fields.Low]  # analysis:ignore

    @property
    def opens(self):
        return self.values[:,Fields.Open]  # analysis:ignore

    @property
    def closes(self):
        return self.values[:,Fields.Close]  # analysis:ignore

    @property
    def volumes(self):
        return self.values[:,Fields.Volume]  # analysis:ignore

    # -------------------------------------------------------------------------
    def curve(self, field="close"):
        """
        Description:
            Convert a HLOCV curve to a scalar curve.
        Inputs:
            field - choose among High, Low, Open, Close, Volume
        Returns:
            A new Curve.
        """
        idx = getattr(Fields, field)
        return self.create_raw(self.dates, self.values[:,idx]) # analysis:ignore



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core.datatypes.date import Date
from onyx.core.database.ufo_base import UfoBase
from onyx.core.database.ufo_fields import FieldError
from onyx.core.database.ufo_fields import StringField, SelectField, IntField
from onyx.core.database.ufo_fields import SelectIntField, FloatField, BoolField
from onyx.core.database.ufo_fields import DateField, SetField

import unittest

__all__ = []


class ufo_class(UfoBase):
    string_attr = StringField()
    string_attr_def = StringField(default="default string")
    select_attr = SelectField(options=["option1", "option2"])
    int_attr = IntField()
    int_attr_def = IntField(default=666)
    int_attr_pos = IntField(positive=True)
    select_int_attr = SelectIntField(options=[1, 2])
    float_attr = FloatField()
    float_attr_def = FloatField(default=666.666)
    float_attr_pos = FloatField(positive=True)
    bool_attr = BoolField()
    date_attr = DateField()
    set_attr = SetField()


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        self.obj = ufo_class()

    def test_string(self):
        self.obj.string_attr = "a string"
        self.assertEqual(self.obj.string_attr_def, "default string")
        with self.assertRaises(FieldError):
            self.obj.string_attr = 123
        with self.assertRaises(FieldError):
            self.obj.string_attr = None

    def test_select(self):
        self.obj.select_attr = "option1"
        self.obj.select_attr = "option2"
        with self.assertRaises(FieldError):
            self.obj.select_attr = 123
        with self.assertRaises(FieldError):
            self.obj.select_attr = None
        with self.assertRaises(FieldError):
            self.obj.select_attr = "option3"

    def test_int(self):
        self.obj.int_attr = 123
        self.assertEqual(self.obj.int_attr_def, 666)
        with self.assertRaises(FieldError):
            self.obj.int_attr = "a string"
        with self.assertRaises(FieldError):
            self.obj.int_attr = 123.456
        with self.assertRaises(FieldError):
            self.obj.int_attr = None
        self.obj.int_attr_pos = 123
        with self.assertRaises(FieldError):
            self.obj.int_attr_pos = -123
        with self.assertRaises(FieldError):
            self.obj.int_attr_pos = 123.456
        with self.assertRaises(FieldError):
            self.obj.int_attr_pos = -123.456

    def test_select_int(self):
        self.obj.select_int_attr = 1
        self.obj.select_int_attr = 2
        with self.assertRaises(FieldError):
            self.obj.select_int_attr = "abc"
        with self.assertRaises(FieldError):
            self.obj.select_int_attr = 1.0
        with self.assertRaises(FieldError):
            self.obj.select_int_attr = None
        with self.assertRaises(FieldError):
            self.obj.select_int_attr = 3

    def test_float(self):
        self.obj.float_attr = 123.456
        self.assertEqual(self.obj.float_attr_def, 666.666)
        with self.assertRaises(FieldError):
            self.obj.float_attr = "a string"
        with self.assertRaises(FieldError):
            self.obj.float_attr = 123
        with self.assertRaises(FieldError):
            self.obj.float_attr = None
        self.obj.float_attr_pos = 123.456
        with self.assertRaises(FieldError):
            self.obj.float_attr_pos = -123.456
        with self.assertRaises(FieldError):
            self.obj.float_attr_pos = 123
        with self.assertRaises(FieldError):
            self.obj.float_attr_pos = -123

    def test_bool(self):
        self.obj.bool_attr = True
        self.obj.bool_attr = False
        with self.assertRaises(FieldError):
            self.obj.bool_attr = 0
        with self.assertRaises(FieldError):
            self.obj.bool_attr = None

    def test_date(self):
        cls = self.obj.__class__
        d = Date.today()
        self.obj.date_attr = d
        with self.assertRaises(FieldError):
            self.obj.date_attr = "abc"
        self.assertEqual(d, cls.date_attr.from_json(cls.date_attr.to_json(d)))

    def test_set(self):
        cls = self.obj.__class__
        s = {1, 2, 3, "a," "b", True}
        self.assertEqual(s, cls.set_attr.from_json(cls.set_attr.to_json(s)))


if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..structure import Structure, StructureError

import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- perform set-up actions, if any
        pass

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_constructors(self):
        struct0 = Structure({"a": 1.0, "b": 2.0, "c": 3.0})
        struct1 = Structure({"a": 1, "b": 2, "c": 3})
        struct2 = Structure(a=1, b=2, c=3)
        struct3 = Structure([("a", 1), ("b", 2), ("c", 3)])
        struct4 = Structure()
        struct4["a"] = 1
        struct4["b"] = 2
        struct4["c"] = 3

        self.assertEqual(struct0, struct1)
        self.assertEqual(struct1, struct2)
        self.assertEqual(struct2, struct3)
        self.assertEqual(struct3, struct4)

    def test_errors(self):
        self.assertRaises(StructureError, Structure, {"a": None})
        self.assertRaises(StructureError, Structure, {"a": "*"})

    def test_add(self):
        struct1 = Structure(a=1, b=2, c=3)
        struct2 = Structure(b=10, c=10, d=10)

        self.assertEqual(1 + struct1, Structure(a=2, b=3, c=4))
        self.assertEqual(struct1 + 1.0, Structure(a=2, b=3, c=4))

        self.assertEqual(struct1 + struct2, Structure(a=1, b=12, c=13, d=10))
        self.assertEqual(struct2 + struct1, Structure(a=1, b=12, c=13, d=10))

        struct1 += struct2
        self.assertEqual(struct1, Structure(a=1, b=12, c=13, d=10))

    def test_sub(self):
        struct1 = Structure(a=1, b=2, c=3)
        struct2 = Structure(b=10, c=10, d=10)

        self.assertEqual(1 - struct1, Structure(a=0, b=-1, c=-2))
        self.assertEqual(struct1 - 1.0, Structure(a=0, b=1, c=2))

        self.assertEqual(struct1 - struct2, Structure(a=1, b=-8, c=-7, d=-10))
        self.assertEqual(struct2 - struct1, Structure(a=-1, b=8, c=7, d=10))

        struct1 -= struct2
        self.assertEqual(struct1, Structure(a=1, b=-8, c=-7, d=-10))

        struct1 -= struct1
        self.assertEqual(struct1.drop_zeros(), Structure())

    def test_mul(self):
        struct = Structure(a=1, b=2, c=3)

        self.assertEqual(2*struct, Structure(a=2, b=4, c=6))
        self.assertEqual(struct*2, Structure(a=2, b=4, c=6))

        struct *= 2.0
        self.assertEqual(struct, Structure(a=2, b=4, c=6))
        struct *= 0.0
        self.assertEqual(struct.drop_zeros(), Structure())

        struct1 = Structure(a=1, b=2, c=3)
        struct2 = Structure(a=1, b=2, c=3)

        self.assertRaises(StructureError, lambda: struct1*struct2)

    def test_pickling(self):
        struct = Structure({"a": 1, "b": 2, "c": 3})
        self.assertEqual(struct, pickle.loads(pickle.dumps(struct, 2)))


if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..date import Date
from ..gcurve import GCurve, Knot, CurveError

import numpy as np
import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- create a first curve by passing arrays of dates and values
        self.dates = [Date.parse(d)
                      for d in ["F08", "G08", "H08", "J08", "K08", "M08",
                                "N08", "Q08", "U08", "V08", "X08", "Z08"]]
        self.values = np.cumsum(np.ones(12))
        self.refcrv = GCurve(self.dates, self.values)

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_len(self):
        self.assertEqual(len(self.refcrv), 12)

    def test_repr(self):
        self.assertEqual(self.refcrv, eval(repr(self.refcrv)))

    def test_constructurs(self):
        # --- test the raw constructor
        crv = GCurve.create_raw(self.refcrv.dates, self.refcrv.values)

        # --- do comparison between curve objects
        self.assertEqual(crv, self.refcrv)

        # --- create a second curve by adding knots iteratively
        crv = GCurve()
        val = 0.0
        for d in self.dates:
            val += 1.0
            crv[d] = val

        # --- do comparison between curve objects
        self.assertEqual(crv, self.refcrv)

        # --- do the comparison of their string representation
        self.assertEqual(crv.__str__(), self.refcrv.__str__())

        # --- check constructor exceptions
        d = Date.today()
        self.assertRaises(CurveError, GCurve, [d, d], [1, 1])
        self.assertRaises(CurveError, GCurve,
                          ([Date(2008, 1, 1), Date(2008, 12, 1)], [1, 1, 1]))

    def test_get(self):
        # --- get an existing knot
        self.assertEqual(self.refcrv[Date(2008, 6, 1)], 6.0)
        # --- test exceptions are raised when the knot is missing
        self.assertRaises(IndexError,
                          self.refcrv.__getitem__, Date(2007, 6, 1))
        self.assertRaises(IndexError,
                          self.refcrv.__getitem__, Date(2008, 6, 15))
        self.assertRaises(IndexError,
                          self.refcrv.__getitem__, Date(2009, 6, 1))

    def test_set(self):
        value = 666
        # --- set at the front of the curve
        date = Date(2007, 1, 1)
        self.refcrv[date] = value
        self.assertEqual(self.refcrv[date], value)
        knot = self.refcrv.front
        self.assertEqual(knot.date, date)
        self.assertEqual(knot.value, value)

        # --- set an existing knot
        date = Date(2008, 6, 1)
        self.assertEqual(self.refcrv[date], 6)
        self.refcrv[date] = value
        self.assertEqual(self.refcrv[date], value)

        # --- set missing knot in the middle
        date = Date(2008, 6, 15)
        self.assertRaises(IndexError, self.refcrv.__getitem__, date)
        self.refcrv[date] = value
        self.assertEqual(self.refcrv[date], value)

        # --- set at the back of the curve
        date = Date.today()
        self.refcrv[date] = value
        self.assertEqual(self.refcrv[date], value)
        knot = self.refcrv.back
        self.assertEqual(knot.date, date)
        self.assertEqual(knot.value, value)

    def test_del(self):
        # --- delete an existing knot
        del self.refcrv[Date(2008, 6, 1)]
        self.assertFalse(Date(2008, 6, 1) in self.refcrv)
        # --- delete a missing knot
        self.assertRaises(IndexError,
                          self.refcrv.__delitem__, Date(2008, 6, 1))
        self.assertRaises(IndexError,
                          self.refcrv.__delitem__, Date(2009, 6, 1))

    def test_has(self):
        self.assertFalse(Date(2007, 6, 1) in self.refcrv)
        self.assertTrue(Date(2008, 6, 1) in self.refcrv)
        self.assertFalse(Date(2008, 6, 15) in self.refcrv)
        self.assertFalse(Date(2009, 6, 1) in self.refcrv)

    def test_back_and_front(self):
        self.assertEqual(self.refcrv.front, Knot(Date(2008, 1, 1), 1.0))
        self.assertEqual(self.refcrv.front.date, Date(2008, 1, 1))
        self.assertEqual(self.refcrv.front.value, 1.0)
        self.assertEqual(self.refcrv.back, Knot(Date(2008, 12, 1), 12.0))
        self.assertEqual(self.refcrv.back.date, Date(2008, 12, 1))
        self.assertEqual(self.refcrv.back.value, 12.0)

    def test_dates_and_values(self):
        values = np.array([1.0, 2.0, 3.0, 4.0, 5.0,
                           6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0])
        self.assertEqual(sum(self.refcrv.values - values), 0.0)
        dates = np.array([733042, 733073, 733102, 733133, 733163, 733194,
                          733224, 733255, 733286, 733316, 733347, 733377])
        self.assertEqual(sum(self.refcrv.ordinals - dates), 0.0)

    def test_pickling(self):
        new_crv = pickle.loads(pickle.dumps(self.refcrv, -1))
        self.assertEqual(new_crv, self.refcrv)

    def test_crop(self):
        sd = Date.parse("N08")
        ed = Date.parse("X08")
        dates = [Date.parse(d) for d in ["N08", "Q08", "U08", "V08", "X08"]]
        values = [7, 8, 9, 10, 11]
        crv = GCurve(dates, values, dtype=float)
        self.assertEqual(self.refcrv.crop(sd, ed), crv)
        self.assertEqual(self.refcrv.crop(Date(2008, 6, 15),
                                          Date(2008, 11, 15)), crv)
        dates = [Date.parse(d) for d in
                 ["N08", "Q08", "U08", "V08", "X08", "Z08"]]
        values = [7, 8, 9, 10, 11, 12]
        crv = GCurve(dates, values, dtype=float)
        self.assertEqual(self.refcrv.crop(Date(2008, 6, 15),
                                          Date(2009, 12, 31)), crv)


if __name__ == "__main__":
    unittest.main()
